
#include <stdio.h>
#include <string.h>
#include "dl.h"
#include "struct.h"
#include "mem.h"
#include <stdlib.h>
#include <syslog.h>
#include <netdb.h>
#include "gc.h"
#include "irc_printf.h"

extern struct oline *olines;
extern struct pline *plines;
extern struct motd *motd_head;

/* void __read_conf_file()
   reads the configuration file
 */

void __read_conf_file()
{
     FILE2 *cfile;
     char line[255];

     memset((void *) server_name, 0, 100);
     memset((void *) server_desc, 0, 150);
     port = 0;
     memset((void *) admin1, 0, 75);
     memset((void *) admin2, 0, 75);
     memset((void *) admin3, 0, 75);

     cfile = irc_fopen("ircd.conf", "r");
     if (cfile == NULL) {
	  fprintf(stderr, "Can't open conf file!\n");
	  exit(1);
     }
     while (!irc_feof(cfile)) {
	  log("beginning loop..");
	  irc_fgets(line, 255, cfile);
	  if (!irc_feof(cfile)) {
	       switch (line[0]) {
	       case '#':
		    break;
	       case 'A':
		    {
			 char *tmp;

			 log("A: line found");
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 strncpy(admin1, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 strncpy(admin2, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 strncpy(admin3, tmp, 75);
		    }
		    break;
	       case 'C':
		    break;
	       case 'D':
		    {
			 char *tmp;

			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 load_module(tmp);
		    }
		    break;
	       case 'd':
		    {
			 char *tmp;

			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 add_ld_path(tmp);
		    }
		    break;
	       case 'F':
		    {
			 char *tmp;
			 struct motd *motd, *list;

			 log("F: line found");
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 motd = (struct motd *) gcAlloc(sizeof(struct motd));
			 bzero(motd, sizeof(struct motd));

			 motd->magic = MOTD_MAGIC;
			 snprintf(motd->hostmask, 75, "%s", tmp);
			 fprintf(stderr, "hostmask: %s\n", motd->hostmask);
			 tmp = strtok(NULL, ":\n");
			 snprintf(motd->motdfile, 75, "%s", tmp);
			 fprintf(stderr, "motdfile: %s\n", motd->motdfile);
			 if (motd_head == NULL) {
			      motd_head = motd;
			 } else {
			      list = motd_head;
			      while (list->next != NULL) {
				   list = list->next;
			      }
			      list->next = motd;
			      motd->prev = list;
			 }
		    }
	       case 'H':
		    break;
	       case 'Y':	/* Y:class:ping freq:conn freq:max clients:sendq */
		    {
			 int class;
			 int ping_freq;
			 int conn_freq;
			 int max_conn;
			 int sendq;
			 struct yline *yline;
			 struct yline *list;
			 char *tmp;
			 yline = (struct yline *) gcAlloc(sizeof(struct yline));
			 bzero(yline, sizeof(struct yline));

			 yline->magic = YLINE_MAGIC;
			 fprintf(stderr, "Y: line found: %s\n", line);
			 tmp = strtok(line, ":\n");	/* Y: */
			 tmp = strtok(NULL, ":\n");	/* class number */
			 class = atoi(tmp);
			 fprintf(stderr, "Y: class = %i\n", class);
			 tmp = strtok(NULL, ":\n");	/* ping frequency */
			 ping_freq = atoi(tmp);
			 fprintf(stderr, "Y: ping freq = %i\n", ping_freq);
			 tmp = strtok(NULL, ":\n");	/* connection frequency */
			 conn_freq = atoi(tmp);
			 fprintf(stderr, "Y: conn freq = %i\n", conn_freq);
			 tmp = strtok(NULL, ":\n");	/* max connections */
			 max_conn = atoi(tmp);
			 fprintf(stderr, "Y: max conn = %i\n", max_conn);
			 tmp = strtok(NULL, ":\n");	/* sendq */
			 sendq = atoi(tmp);
			 yline->class = class;
			 yline->ping_freq = ping_freq;
			 yline->conn_freq = conn_freq;
			 yline->max_clients = max_conn;
			 yline->num_connects = 0;
			 yline->sendq = sendq;
			 list = yline_list;
			 yline->refcount = 1;
			 if (!list) {
			      yline_list = yline;
			 } else {
			      int ok = 1;

			      log("checking for already defined Y: line");
			      while (list && ok) {
				   if (list->class == yline->class) {
					log("found one, freeing this one");
					gcDecRefcnt(yline);
					ok = 0;
				   }
				   list = list->next;
			      }
			      if (ok) {
				   log("none found, adding this one to the list");
				   list = yline_list;
				   while (list->next)
					list = list->next;
				   list->next = yline;
				   yline->prev = list;
			      }
			 }
		    }
		    break;
	       case 'K':
		    break;
	       case 'M':
		    {
			 char *tmp;

			 log("M: line found");
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 strncpy(server_name, tmp, 100);
			 tmp = strtok(NULL, ":\n");
			 strncpy(server_hn, tmp, 100);
			 tmp = strtok(NULL, ":\n");
			 strncpy(server_desc, tmp, 150);
			 tmp = strtok(NULL, ":\n");
			 port = strtol(tmp, (char **) NULL, 10);
		    }
		    break;
	       case 'N':
		    log("entered N: line code");
		    break;
	       case 'O':
		    {
			 char *tmp;
			 struct oline *new_ent;

			 log("O: line found");
			 new_ent = (struct oline *) gcAlloc(sizeof(struct oline));

			 new_ent->magic = OLINE_MAGIC;
			 log("allocated struct for oline");
			 {
			      if (olines == NULL) {
				   log("olines is null..");
				   olines = new_ent;
			      } else {
				   struct oline *lines;

				   lines = olines;
				   while (lines->next != NULL)
					lines = lines->next;
				   lines->next = new_ent;
				   new_ent->prev = lines;
			      }
			 }
			 log("parsing O: line");
			 tmp = strtok(line, ":\n");
			 log("skipped O:");
			 tmp = strtok(NULL, ":\n");
			 {
			      char t[255];

			      snprintf(t, 255, "hostmask found: %s", tmp);
			      log(t);
			 }
			 strncpy(new_ent->hostmask, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 {
			      char t[255];

			      snprintf(t, 255, "password found: %s", tmp);
			      log(t);
			 }
			 strncpy(new_ent->pass, tmp, 15);
			 tmp = strtok(NULL, ":\n");
			 {
			      char t[255];

			      snprintf(t, 255, "nickname found: %s", tmp);
			      log(t);
			 }
			 strncpy(new_ent->nick, tmp, 15);
			 tmp = strtok(NULL, ":\n");
			 tmp = strtok(NULL, ":\n");
			 new_ent->class = strtol(tmp, (char **) NULL, 10);
		    }
		    break;
	       case 'P':
		    {
			 char *tmp;
			 struct pline *new_ent;

			 fprintf(stderr, "line is '%s'\n", line);

			 tmp = strtok(line, ":\n");	/* P: */
			 tmp = strtok(NULL, ":\n");	/* <host> */

			 fprintf(stderr, "value is '%s'\n", tmp);
			 new_ent = (struct pline *) gcAlloc(sizeof(struct pline));
			 bzero(new_ent, sizeof(struct pline));

			 new_ent->magic = PLINE_MAGIC;
			 snprintf(new_ent->hostname, 60, "%s", tmp);
			 tmp = strtok(NULL, ":\n");	/* "*" */
			 tmp = strtok(NULL, ":\n");	/* "*" */
			 tmp = strtok(NULL, ":\n");	/* port num */
			 new_ent->port = strtol(tmp, (char **) 0, 10);
			 if (new_ent->hostname[0] != '*') {
			      struct hostent *host_struct, he_buf;
			      char he_buf2[100];

			      gethostbyname_r(new_ent->hostname, &he_buf, he_buf2, 100, &host_struct, &h_errno);
			      if (host_struct != NULL) {
				   int *i = (int *) host_struct->h_addr_list[0];

				   new_ent->ipnum = *i;
			      } else {
				   new_ent->ipnum = 0;
			      }
			 }
			 if (plines == NULL) {
			      plines = new_ent;
			 } else {
			      struct pline *list;

			      list = plines;
			      while (list->next != NULL)
				   list = list->next;
			      list->next = new_ent;
			      new_ent->prev = list;
			 }
		    }
		    break;
	       case 'Q':
		    break;
	       case 'T':
		    break;
	       case 'U':
		    break;
	       case 'I':	/* I:ip mask:pass:host mask:*:class */
		    {
			 char ip_mask[20];
			 char pass[20];
			 char host_mask[75];
			 int class;
			 struct yline *y_class;
			 struct iline *iline;
			 char *tmp;

			 bzero(ip_mask, 20);
			 bzero(pass, 20);
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 strncpy(ip_mask, tmp, 20);
			 tmp = strtok(NULL, ":\n");
			 strncpy(pass, tmp, 20);
			 tmp = strtok(NULL, ":\n");
			 strncpy(host_mask, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 tmp = strtok(NULL, ":\n");
			 class = atoi(tmp);
			 y_class = yline_list;
			 iline = (struct iline *) gcAlloc(sizeof(struct iline));
			 bzero(iline, sizeof(struct iline));

			 iline->magic = ILINE_MAGIC;

			 while (y_class) {
			      if (y_class->class == class) {
				   gcIncRefcnt(y_class);
				   iline->class = y_class;
				   y_class->refcount++;
				   y_class = NULL;
			      }
			      if (y_class)
				   y_class = y_class->next;
			 }
			 if (!iline->class) {
			      fprintf(stderr, "Bad class for this iline! Did you define the Y: "
				      "first? iline class = %i\n", class);
			      gcDecRefcnt(iline);
			 } else {
			      struct iline *list;

			      strcpy(iline->ip_addr_mask, ip_mask);
			      strcpy(iline->password, pass);
			      strcpy(iline->hostmask, host_mask);
			      list = iline_list;
			      if (list) {
				   while (list->next)
					list = list->next;
				   list->next = iline;
				   iline->prev = list;
			      } else {
				   iline_list = iline;
			      }
			 }
		    }
		    break;
	       case 'Z':
		    break;
	       default:
	       };
	  }
     }
     if (server_name[0] == 0) {
	  fprintf(stderr, "No M: line in conf file!\n");
	  exit(1);
     }
     irc_fclose(cfile);
}

extern LD_Path *ld_path_list;
extern Module *module_list;

/* m_rehash
   av[0] = "REHASH"
   reloads the config file
 */
void m_rehash(struct fd_list *fd, char **av, int ac)
{
     FILE2 *cfile;
     char buf[255];

     syslog(LOG_CRIT, "%s is rehashing ircd.conf!", fd->nick);

     cfile = irc_fopen("ircd.conf", "r");
     if (!cfile) {
	  irc_fprintf(fd->sok, ":%s 424 %s :File error doing fopen on ircd.conf\r\n", server_name, fd->nick);
	  return;
     }
     /* since we're re-reading the conf file, clean out the current settings, except for M: lines */

     {
	  struct pline *p;

	  p = plines;
	  log("cleaning p: lines");
	  while (p) {
	       struct pline *q = p->next;

	       close(p->fd);
	       gcDecRefcnt(p);
	       p = q;
	  }
	  log("done");
	  plines = NULL;
     }

     {
	  struct oline *o;

	  o = olines;
	  log("cleaning o: lines");
	  while (o) {
	       struct oline *n = o->next;

	       gcDecRefcnt(o);
	       o = n;
	  }
	  log("done");
	  olines = NULL;
     }

     {
	  LD_Path *path = ld_path_list->next;

	  log("cleaning d: lines");
	  while (path) {
	       LD_Path *q = path->next;

	       gcDecRefcnt(path);
	       path = q;
	  }
	  ld_path_list->next = NULL;
	  log("done");
     }

     {
	  Module *list = module_list->next;
	  char **modules = NULL;
	  int nmodules = 0;
	  int i;

	  log("freeing modules");
	  while (list) {
	       modules = realloc(modules, ++nmodules * sizeof(char *));

	       modules[nmodules - 1] = malloc(strlen(list->info->module_name));
	       strcpy(modules[nmodules - 1], list->info->module_name);
	       list = list->next;
	  }
	  for (i = 0; i < nmodules; i++) {
	       unload_module(modules[i]);
	       free(modules[i]);
	  }
	  module_list->next = NULL;
	  log("done");
	  free(modules);
     }

     {
	  struct iline *i = iline_list;

	  log("freeing I: lines");
	  while (i) {
	       struct iline *q = i->next;

	       i->class->refcount--;
	       if (i->class->refcount == 0) {
		    gcDecRefcnt(i->class);
	       }
	       gcDecRefcnt(i);
	       i = q;
	  }
	  iline_list = NULL;
     }

     {
	  struct yline *y = yline_list;

	  log("freeing Y: lines");
	  while (y) {
	       struct yline *q = y->next;

	       y->refcount--;
	       if (y->refcount == 0) {
		    gcDecRefcnt(y);
	       } else {
		    y->prev = y->next = NULL;
	       }
	       y = q;
	  }
	  yline_list = NULL;
     }

     {
	  struct motd *m = motd_head;

	  log("freeing F: lines");
	  while (m) {
	       struct motd *q = m->next;

	       gcDecRefcnt(m);
	       m = q;
	  }
	  motd_head = NULL;
     }

     while (!irc_feof(cfile)) {
	  irc_fgets(buf, 255, cfile);
	  if (!irc_feof(cfile)) {
	       char *line = buf;

	       log(buf);

	       switch (buf[0]) {
	       case 'A':
		    {
			 char *tmp;

			 log("A: line found");
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 strncpy(admin1, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 strncpy(admin2, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 strncpy(admin3, tmp, 75);
		    }
		    break;
	       case 'D':
		    {
			 char *tmp;

			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 load_module(tmp);
		    }
		    break;
	       case 'd':
		    {
			 char *tmp;

			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 add_ld_path(tmp);
		    }
		    break;
	       case 'F':
		    {
			 char *tmp;
			 struct motd *motd, *list;

			 log("F: line found");
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 motd = (struct motd *) gcAlloc(sizeof(struct motd));
			 bzero(motd, sizeof(struct motd));

			 motd->magic = MOTD_MAGIC;
			 snprintf(motd->hostmask, 75, "%s", tmp);
			 fprintf(stderr, "hostmask: %s\n", motd->hostmask);
			 tmp = strtok(NULL, ":\n");
			 snprintf(motd->motdfile, 75, "%s", tmp);
			 fprintf(stderr, "motdfile: %s\n", motd->motdfile);
			 if (motd_head == NULL) {
			      motd_head = motd;
			 } else {
			      list = motd_head;
			      while (list->next != NULL) {
				   list = list->next;
			      }
			      list->next = motd;
			      motd->prev = list;
			 }
		    }
		    break;
	       case 'I':
		    {
			 char ip_mask[20];
			 char pass[20];
			 char host_mask[75];
			 int class;
			 struct yline *y_class;
			 struct iline *iline;
			 char *tmp;

			 bzero(ip_mask, 20);
			 bzero(pass, 20);
			 tmp = strtok(line, ":\n");
			 tmp = strtok(NULL, ":\n");
			 strncpy(ip_mask, tmp, 20);
			 tmp = strtok(NULL, ":\n");
			 strncpy(pass, tmp, 20);
			 tmp = strtok(NULL, ":\n");
			 strncpy(host_mask, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 tmp = strtok(NULL, ":\n");
			 class = atoi(tmp);
			 y_class = yline_list;
			 iline = (struct iline *) gcAlloc(sizeof(struct iline));
			 bzero(iline, sizeof(struct iline));

			 iline->magic = ILINE_MAGIC;
			 while (y_class) {
			      if (y_class->class == class) {
				   log("increasing Y: line refcount..");
				   gcIncRefcnt(y_class);
				   iline->class = y_class;
				   y_class->refcount++;
				   log("done");
				   y_class = NULL;
			      }
			      if (y_class)
				   y_class = y_class->next;
			 }
			 if (!iline->class) {
			      fprintf(stderr, "Bad class for this iline! Did you define the Y: "
				      "first? iline class = %i\n", class);
			      gcDecRefcnt(iline);
			 } else {
			      struct iline *list;

			      log("adding iline to the list");
			      log("-- copying ip_mask to iline->ip_addr_mask");
			      strcpy(iline->ip_addr_mask, ip_mask);
			      log("-- copying pass to iline->password");
			      strcpy(iline->password, pass);
			      log("-- copying host_mask to iline->hostmask");
			      strcpy(iline->hostmask, host_mask);
			      log("-- traversing linked list");
			      list = iline_list;
			      if (list) {
				   while (list->next)
					list = list->next;
				   list->next = iline;
				   iline->prev = list;
			      } else {
				   iline_list = iline;
			      }
			      log("done");
			 }
			 log("finished adding I: line");
		    }
		    break;
	       case 'O':
		    {
			 char *tmp;
			 struct oline *new_ent;

			 log("O: line found");
			 new_ent = (struct oline *) gcAlloc(sizeof(struct oline));

			 new_ent->magic = OLINE_MAGIC;
			 log("allocated struct for oline");
			 {
			      if (olines == NULL) {
				   log("olines is null..");
				   olines = new_ent;
			      } else {
				   struct oline *lines;

				   lines = olines;
				   while (lines->next != NULL)
					lines = lines->next;
				   lines->next = new_ent;
				   new_ent->prev = lines;
			      }
			 }
			 log("parsing O: line");
			 tmp = strtok(line, ":\n");
			 log("skipped O:");
			 tmp = strtok(NULL, ":\n");
			 {
			      char t[255];

			      snprintf(t, 255, "hostmask found: %s", tmp);
			      log(t);
			 }
			 strncpy(new_ent->hostmask, tmp, 75);
			 tmp = strtok(NULL, ":\n");
			 {
			      char t[255];

			      snprintf(t, 255, "password found: %s", tmp);
			      log(t);
			 }
			 strncpy(new_ent->pass, tmp, 15);
			 tmp = strtok(NULL, ":\n");
			 {
			      char t[255];

			      snprintf(t, 255, "nickname found: %s", tmp);
			      log(t);
			 }
			 strncpy(new_ent->nick, tmp, 15);
		    }
		    break;
	       case 'P':
		    {
			 char *tmp;
			 struct pline *new_ent;

			 fprintf(stderr, "line is '%s'\n", line);

			 tmp = strtok(line, ":\n");	/* P: */
			 tmp = strtok(NULL, ":\n");	/* <host> */

			 fprintf(stderr, "value is '%s'\n", tmp);
			 new_ent = (struct pline *) gcAlloc(sizeof(struct pline));
			 bzero(new_ent, sizeof(struct pline));

			 new_ent->magic = PLINE_MAGIC;
			 snprintf(new_ent->hostname, 60, "%s", tmp);
			 tmp = strtok(NULL, ":\n");	/* "*" */
			 tmp = strtok(NULL, ":\n");	/* "*" */
			 tmp = strtok(NULL, ":\n");	/* port num */
			 new_ent->port = strtol(tmp, (char **) 0, 10);
			 if (new_ent->hostname[0] != '*') {
			      struct hostent *host_struct, he_buf;
			      char he_buf2[100];

			      gethostbyname_r(new_ent->hostname, &he_buf, he_buf2, 100, &host_struct, &h_errno);
			      if (host_struct != NULL) {
				   int *i = (int *) host_struct->h_addr_list[0];

				   new_ent->ipnum = *i;
			      } else {
				   new_ent->ipnum = 0;
			      }
			 }
			 if (plines == NULL) {
			      plines = new_ent;
			 } else {
			      struct pline *list;

			      list = plines;
			      while (list->next != NULL)
				   list = list->next;
			      list->next = new_ent;
			      new_ent->prev = list;
			 }

		    }
		    break;
	       case 'Y':	/* Y:class:ping freq:conn freq:max clients:sendq */
		    {
			 int class;
			 int ping_freq;
			 int conn_freq;
			 int max_conn;
			 int sendq;
			 struct yline *yline;
			 struct yline *list;
			 char *tmp;
			 yline = (struct yline *) gcAlloc(sizeof(struct yline));
			 bzero(yline, sizeof(struct yline));

			 yline->magic = YLINE_MAGIC;
			 fprintf(stderr, "Y: line found: %s\n", line);
			 tmp = strtok(line, ":\n");	/* Y: */
			 tmp = strtok(NULL, ":\n");	/* class number */
			 class = atoi(tmp);
			 fprintf(stderr, "Y: class = %i\n", class);
			 tmp = strtok(NULL, ":\n");	/* ping frequency */
			 ping_freq = atoi(tmp);
			 fprintf(stderr, "Y: ping freq = %i\n", ping_freq);
			 tmp = strtok(NULL, ":\n");	/* connection frequency */
			 conn_freq = atoi(tmp);
			 fprintf(stderr, "Y: conn freq = %i\n", conn_freq);
			 tmp = strtok(NULL, ":\n");	/* max connections */
			 max_conn = atoi(tmp);
			 fprintf(stderr, "Y: max conn = %i\n", max_conn);
			 tmp = strtok(NULL, ":\n");	/* sendq */
			 sendq = atoi(tmp);
			 yline->class = class;
			 yline->ping_freq = ping_freq;
			 yline->conn_freq = conn_freq;
			 yline->max_clients = max_conn;
			 yline->num_connects = 0;
			 yline->sendq = sendq;
			 list = yline_list;
			 yline->refcount = 1;
			 if (!list) {
			      yline_list = yline;
			 } else {
			      int ok = 1;

			      while (list) {
				   if (list->class == yline->class) {
					gcDecRefcnt(yline);
					ok = 0;
				   }
				   list = list->next;
			      }
			      if (ok) {
				   list = yline_list;
				   while (list->next)
					list = list->next;
				   list->next = yline;
				   yline->prev = list;
			      }
			 }
		    }
		    break;
	       default:
		    break;
	       }
	  }
     }

     {
	  struct pline *q;

	  q = plines;
	  while (q) {
	       q->fd = open_listen_sok_2(q->port, q->ipnum);
	       q = q->next;
	  }
     }

     irc_fprintf(fd->sok, ":%s 382 %s %s :Rehashing\r\n", server_name, fd->nick, "ircd.conf");
     irc_fclose(cfile);
     return;
}
