
#include <stdio.h>
#include <pthread.h>
#include <errno.h>

#include "gc.h"

/*
   struct file_struct {
   FILE *sok;
   pthread_mutex_t mut;
   };

   typedef struct file_struct FILE2;
 */

extern int errno;

pthread_mutex_t stdout_mut = PTHREAD_MUTEX_INITIALIZER;

FILE2 *irc_fopen(char *path, char *mode)
{
     FILE *file;
     FILE2 *f2buf;

     file = fopen(path, mode);
     if (!file) {
	  return NULL;
     } else {
	  f2buf = gcAlloc(sizeof(FILE2));
	  pthread_mutex_init(&(f2buf->mut), NULL);
	  f2buf->sok = file;
	  return f2buf;
     }
}

FILE2 *irc_fdopen(int fd, char *mode)
{
     FILE *file;
     FILE2 *f2buf;

     file = fdopen(fd, mode);
     if (!file) {
	  return NULL;
     } else {
	  f2buf = gcAlloc(sizeof(FILE2));
	  pthread_mutex_init(&(f2buf->mut), NULL);
	  f2buf->sok = file;
	  return f2buf;
     }
}

int irc_fclose(FILE2 * str)
{
     if (!str) {
	  errno = EBADF;
	  return EOF;
     }
     pthread_mutex_lock(&str->mut);
     fclose(str->sok);
     str->sok = NULL;
     pthread_mutex_unlock(&(str->mut));
     pthread_mutex_destroy(&(str->mut));
     gcDecRefcnt(str);
     return 0;
}

int irc_printf(const char *format,...) __attribute__((__format__(printf, 1, 2)));

int irc_printf(const char *format,...)
{
     va_list ap;
     int l;

     pthread_mutex_lock(&stdout_mut);
     va_start(ap, format);
     l = vfprintf(stdout, format, ap);
     va_end(ap);
     pthread_mutex_unlock(&stdout_mut);

     return l;
}

int irc_fprintf(FILE2 * str, const char *fmt,...) __attribute__((__format__(printf, 2, 3)));

int irc_fprintf(FILE2 * str, const char *fmt,...)
{
     va_list ap;
     int l;

     if (!str) {
	  errno = EINVAL;
	  return -1;
     }
     pthread_mutex_lock(&(str->mut));
     va_start(ap, fmt);
     if (str->sok) {
	  l = vfprintf(str->sok, fmt, ap);
     } else {
	  l = 0;
     }
     va_end(ap);
     pthread_mutex_unlock(&(str->mut));

     return l;
}

int irc_vprintf(const char *fmt, va_list ap)
{
     int l;

     pthread_mutex_lock(&stdout_mut);
     l = vfprintf(stdout, fmt, ap);
     pthread_mutex_unlock(&stdout_mut);

     return l;
}

int irc_vfprintf(FILE2 * str, const char *fmt, va_list ap)
{
     int l;

     if (!str) {
	  errno = EINVAL;
	  return -1;
     }
     pthread_mutex_lock(&(str->mut));
     if (str->sok) {
	  l = vfprintf(str->sok, fmt, ap);
     } else {
	  l = 0;
     }
     pthread_mutex_unlock(&(str->mut));

     return l;
}

int irc_fflush(FILE2 * str)
{
     pthread_mutex_lock(&(str->mut));
     fflush(str->sok);
     pthread_mutex_unlock(&(str->mut));
     return 0;
}

int irc_feof(FILE2 * str)
{
     int l;

     pthread_mutex_lock(&(str->mut));
     l = feof(str->sok);
     pthread_mutex_unlock(&(str->mut));
     return l;
}

int irc_fgetc(FILE2 * str)
{
     int c;

     pthread_mutex_lock(&(str->mut));
     if (str->sok) {
	  c = fgetc(str->sok);
     } else {
	  c = 0;
     }
     pthread_mutex_unlock(&(str->mut));
     return c;
}

char *irc_fgets(char *s, int len, FILE2 * str)
{
     pthread_mutex_lock(&(str->mut));
     if (str->sok) {
	  fgets(s, len, str->sok);
     } else {
	  s[0] = 0;
     }
     pthread_mutex_unlock(&(str->mut));
     return s;
}

int irc_getchar()
{
     FILE2 sok;
     int i;

     sok.sok = stdin;
     pthread_mutex_init(&(sok.mut), NULL);
     i = irc_fgetc(&sok);
     pthread_mutex_destroy(&(sok.mut));
     return i;
}

/* this is a very dangerous function to use - so i just return null. */

char *irc_gets(char *s)
{
     return NULL;
}

int irc_ungetc(int c, FILE2 * str)
{
     int l;

     pthread_mutex_lock(&(str->mut));
     if (str->sok) {
	  l = ungetc(c, str->sok);
     } else {
	  l = EOF;
     }
     pthread_mutex_unlock(&(str->mut));
     return l;
}
