
#ifndef __struct_h__
#define __struct_h__

#define IS_CLIENT 1;
#define IS_SERVER 2;
#define IS_UNKNOWN 3;

#include <pthread.h>

#include "irc_printf.h"

#define FD_MAGIC 0xbcdef012

struct fd_list {
     int magic;
     struct fd_list *next;
     struct fd_list *prev;
     FILE2 *sok;
     int fd;
     int is_server;
     char nick[15];
     char userid[15];
     char hostname[50];
     char server_name[50];
     char realname[100];
     char conn_pass[20];
     int registered;
     int signon;
     int last_msg;
     char invisible;
     char s_msg;
     char wallops;
     char oper;
     char deaf;
     char service;
     char globops;
     char show_whois;
     char show_connect;
     char hide_hostname;
     char buf[1024];
     int away_set;
     char away_msg[150];
     void *chan_list;
     void *invite_list;
     struct server_list *sptr;
     int last_ping;
     int last_ping_reply;
     struct {
	  int time;
	  int nmsgs;
     } msgs;
     int ident_is_in_use;
     int nslookup_is_in_use;
     int waiting_for_reg;
#ifdef PTHREADIDENT
     pthread_t ident_id;
     pthread_t nslookup_id;
     pthread_t reg_id;
#endif
     int is_killed;
     int ref_cnt;
     struct yline *class;
     struct {
	  int length;
	  char *buf;
	  pthread_mutex_t mut;
     } sq;
};

#define SERVER_MAGIC 0xcdef0123
struct server_list {
     int magic;
     struct server_list *prev;
     struct server_list *next;
     FILE *sok;
     int fd;
     char server_name[100];
     char server_desc[150];
     struct fd_list *clients;
     int hops;
};

#define USER_LIST_MAGIC 0xdef01234

struct user_list {
     int magic;
     struct user_list *prev;
     struct user_list *next;
     struct fd_list *user;
     char oper;
     char voice;
};

#define INVITE_LIST_MAGIC 0xef012345

struct invite_list {
     int magic;
     struct invite_list *prev;
     struct invite_list *next;
     struct fd_list *user;
};

#define BAN_LIST_MAGIC 0xf0123456

struct ban_list {
     int magic;
     struct ban_list *prev;
     struct ban_list *next;
     char banid[75];
     int time;
     char userid[120];
};

#define CHAN_LIST_MAGIC 0x01234567

struct chan_list {
     int magic;
     struct chan_list *prev;
     struct chan_list *next;
     char name[200];
     int chan_created_at;
     char topic[1024];
     char topic_set_by[15];
     int topic_set_at;
     struct user_list *users;
     char oper_only_topic;
     char no_outside_msg;
     char private;
     char secret;
     char invite_only;
     char mute;
     char key;
     char limit;
     struct invite_list *invites;
     struct ban_list *bans;
     struct ban_list *exempts;
     struct ban_list *perminvites;
     char key_val[15];
     int limit_val;
};

#define USER_CHAN_LIST_MAGIC 0x12345678

struct user_chan_list {
     int magic;
     struct user_chan_list *prev;
     struct user_chan_list *next;
     struct chan_list *chan;
};

#define USER_INVITE_LIST_MAGIC 0x23456789

struct user_invite_list {
     int magic;
     struct user_invite_list *prev;
     struct user_invite_list *next;
     struct chan_list *chan;
};

#define OLINE_MAGIC 0x3456789a

struct oline {
     int magic;
     struct oline *prev;
     struct oline *next;
     char hostmask[75];
     char nick[15];
     char pass[15];
     int class;
};

#define PLINE_MAGIC 0x456789ab

struct pline {
     int magic;
     struct pline *prev;
     struct pline *next;
     char hostname[60];
     int ipnum;
     int port;
     int fd;
};

#define YLINE_MAGIC 0x56789abc

struct yline {
     int magic;
     struct yline *prev;
     struct yline *next;
     int class;
     int ping_freq;
     int conn_freq;
     int max_clients;
     int num_connects;
     int sendq;
     int refcount;
};

#define ILINE_MAGIC 0x6789abcd

struct iline {
     int magic;
     struct iline *prev;
     struct iline *next;
     struct yline *class;
     char ip_addr_mask[20];
     char password[20];
     char hostmask[75];
};

#define MOTD_MAGIC 0x789abcde

struct motd {
     int magic;
     struct motd *prev, *next;
     char hostmask[75];
     char motdfile[75];
};

extern char server_name[100];
extern int port;
extern char server_desc[150];
extern char server_hn[100];
extern int server_ip;

extern char admin1[75];
extern char admin2[75];
extern char admin3[75];

extern struct chan_list *__clist_head;

extern struct fd_list *current_socket;
extern struct fd_list *__head;

extern struct iline *iline_list;

extern int errno;

struct mchg {
     char orig, new;
};

struct user_mode_change {
     struct mchg invisible, s_msg, wallops, oper, deaf, service, globops, show_whois, show_connect,
       hide_hostname;
     int num_add;
     int num_sub;
};

struct chan_mode_change {
     struct mchg private, secret, invite_only, oper_only_topic, no_outside_msg, mute;
     struct {
	  int orig;
	  int new;
	  char *val;
     } key;
     struct {
	  int orig;
	  int new;
	  int orig_val;
	  int new_val;
     } limit;
};

#define WHOWAS_MAGIC 0x89abcdef

struct whowas {
     int magic;
     struct whowas *prev;
     struct whowas *next;
     char nick[15];
     char userid[15];
     char hostname[50];
     char realname[100];
     int signoff;
};

extern struct whowas *__whowas_head;
extern struct yline *yline_list;

extern int start_time;

#undef char

#endif /* !__struct_h__ */
